/***************************************************************************
***
***    Copyright 2006  Hon Hai Precision Ind. Co. Ltd.
***    All Rights Reserved.
***    No portions of this material shall be reproduced in any form without the
***    written permission of Hon Hai Precision Ind. Co. Ltd.
***
***    All information contained in this document is Hon Hai Precision Ind.  
***    Co. Ltd. company private, proprietary, and trade secret property and 
***    are protected by international intellectual property laws and treaties.
***
****************************************************************************/
 
#include <common.h>

#include "board_config.h"
#include "lzma.h"
#include "mvGpp.h"

/* #define _DEBUG */

/* GPIO definitions */
#define GPIO_RESET          3
#define GPIO_LED_GREEN      0
#define GPIO_LED_AMBER      5
#define GPIO_LED_WLAN       11
#define GPIO_LED_WAN        31

typedef struct partition_S
{
    char name[20];
    unsigned long start_addr;
    unsigned long size;
} partition_t;

/* Partition map of WNR894T, it need to compatible with /linux/drivers/maps/physmap.c */
static partition_t partitions[] = 
{
    { "kernel",     0xFF800000, 0x100000 },     /* 1024 KB */
    { "rootfs",     0xFF900000, 0x630000 },     /* 6336 KB */
    { "T_meter1",   0xFFF30000, 0x10000  },     /*   64 KB */
    { "T_meter2",   0xFFF40000, 0x10000  },     /*   64 KB */
    { "POT",        0xFFF50000, 0x10000  },     /*   64 KB */
    { "board_data", 0xFFF60000, 0x10000  },     /*   64 KB */
    { "nvram",      0xFFF70000, 0x10000  },     /*   64 KB */
    { "u-boot",     0xFFF80000, 0x80000  },     /*  512 KB */
};

#define swap32(val) \
	((unsigned int)( \
		(((unsigned int)(val) & (unsigned int)0x000000ffUL) << 24) | \
		(((unsigned int)(val) & (unsigned int)0x0000ff00UL) <<  8) | \
		(((unsigned int)(val) & (unsigned int)0x00ff0000UL) >>  8) | \
		(((unsigned int)(val) & (unsigned int)0xff000000UL) >> 24) ))

/* Image header format */
typedef struct image_hdr_S
{
    unsigned long magic;            /* magic */
    unsigned long header_len;       /* Length of header */
    unsigned char reserved[8];
    unsigned long kernel_chksum;    /* Kernel image chksum */
    unsigned long rootfs_chksum;    /* rootfs image chksum */
    unsigned long kernel_len;       /* Length of kernel */
    unsigned long rootfs_len;       /* Length of rootfs */
    unsigned long image_chksum;     /* checksum across length of image */
    unsigned long header_chksum;    /* checksum across length of header */
} image_hdr_t;

static 
unsigned long calculate_checksum (int action, char *s, int size)
{
    static unsigned long c0, c1;
    unsigned long checksum, b;
    int i;

    switch (action)
    {
        case 0:
            c0 = c1 = 0;
            break;

        case 1:
            for (i = 0; i < size; i++)
            {
                c0 += s[i] & 0xff;
                c1 += c0;
            }
            break;
    
        case 2:
            b = (c0 & 65535) + ((c0 >> 16) & 65535);
            c0 = ((b >> 16) + b) & 65535;

            b = (c1 & 65535) + ((c1 >> 16) & 65535);
            c1 = ((b >> 16) + b) & 65535;

            checksum = ((c1 << 16) | c0);

            return checksum;
    }
    return 0;
}

static
int verify_checksum(char *buf, unsigned long buf_len, unsigned long chksum)
{
    unsigned long cal_chksum = 0;

    calculate_checksum (0, NULL, 0);
    calculate_checksum (1, buf, buf_len);
    cal_chksum = calculate_checksum (2, NULL, 0); 

    if (cal_chksum != chksum) {
        printf("Checksum mismatch: \n");
        printf("Image chksum: 0x%08X\n", chksum);
        printf("Calc  chksum: 0x%08X\n", cal_chksum);
        return -1;
    }

    return 0;   
}

static
int verify_board_id(char *buf)
{
    unsigned long bd_addr = get_part_addr(PART_BD);
    struct board_data *bd = (struct board_data *)bd_addr;

//    if (strcmp(buf, bd->board_id) != 0)
    if (strcmp(buf, "U12H092T01_NETGEAR") != 0)
    {
        printf("Board ID mismatch!\n");
        printf("Board ID : %s\n", bd->board_id);
        printf("Image ID : %s\n", buf);
        return -1;
    }

    return 0;
}

unsigned long get_part_addr(int part_num)
{
    return partitions[part_num].start_addr;
}

unsigned long get_part_size(int part_num)
{
    return partitions[part_num].size;
}

int get_eth_mac(unsigned char *lan_mac, unsigned char *wan_mac)
{
    unsigned long bd_addr = get_part_addr(PART_BD);
    struct board_data *bd = (struct board_data *)bd_addr;

    /* Read LAN MAC */
    memcpy(lan_mac, bd->lan_mac, 6);

    /* Sanity check for multicast address, and all-0 address */
    if (lan_mac[0] & 0x01)
        lan_mac[0] = 0x00;

    if (memcmp(lan_mac, "\0\0\0\0\0\0", 6) == 0)
        lan_mac[5] = 0x01;

    /* Read WAN MAC */
    memcpy(wan_mac, bd->wan_mac, 6);

    /* Sanity check for multicast address, and all-0 address */
    if (wan_mac[0] & 0x01)
        wan_mac[0] = 0x00;

    if (memcmp(wan_mac, "\0\0\0\0\0\0", 6) == 0)
        wan_mac[5] = 0x02;

    return 0;
}

int verify_flash_image(void)
{
    unsigned long kernel_start, kernel_end, kernel_len, kernel_chksum;
    unsigned long rootfs_start, rootfs_end, rootfs_len, rootfs_chksum;
    int ret = 0;

    kernel_start  = get_part_addr(PART_KERNEL);
    kernel_end    = kernel_start + get_part_size(PART_KERNEL);
    kernel_len    = *((unsigned long *)(kernel_end - 8));
    kernel_chksum = *((unsigned long *)(kernel_end - 4));

    rootfs_start  = get_part_addr(PART_ROOTFS);
    rootfs_end    = rootfs_start + get_part_size(PART_ROOTFS);
    rootfs_len    = *((unsigned long *)(rootfs_end - 8));
    rootfs_chksum = *((unsigned long *)(rootfs_end - 4));
    
    /* Sanity check */
    if ((kernel_len > get_part_size(PART_KERNEL)) ||
        (rootfs_len > get_part_size(PART_ROOTFS)))
    {
        printf("Valid kernel / rootfs len\n");
        return -1;
    }
    
    if (verify_checksum((char *)kernel_start, kernel_len, kernel_chksum))
    {
        printf("kernel checksum invalid\n");
        ret = -1;
    }

    if (verify_checksum((char *)rootfs_start, rootfs_len, rootfs_chksum))
    {
        printf("rootfs checksum invalid\n");
        ret =  -1;
    }

    return ret;
}

int decode_flash_image(void)
{
    unsigned long kernel_start = get_part_addr(PART_KERNEL);
    unsigned long kernel_size  = get_part_size(PART_KERNEL);

    unsigned char *src_addr, *dst_addr;
    unsigned int  src_len, dst_len;
    
    src_addr = (unsigned char *)kernel_start;
    src_len  = kernel_size;
    dst_addr = (unsigned char *)KERNEL_RAM_ADDR;

    printf("Decompressing kernel...\n");
    return lzma_decode(src_addr, dst_addr, src_len, &dst_len);
}

int verify_download_image(unsigned long image_addr)
{
    unsigned char *image_buf = (unsigned char *)image_addr;
    image_hdr_t *header      = (image_hdr_t *)image_buf;
    unsigned long header_len, board_id_len;
    unsigned char board_id[BOARD_ID_LEN];
    unsigned long kernel_len, kernel_chksum;
    unsigned long rootfs_len, rootfs_chksum;
    
    header_len    = swap32(header->header_len);

    kernel_len    = swap32(header->kernel_len);
    kernel_chksum = swap32(header->kernel_chksum);

    rootfs_len    = swap32(header->rootfs_len);
    rootfs_chksum = swap32(header->rootfs_chksum);
    
    /* Sanity Check */
    if (kernel_len > get_part_addr(PART_KERNEL) ||
        rootfs_len > get_part_addr(PART_ROOTFS))
        return -1;
    
    /* Check Board ID first */
    board_id_len = header_len - sizeof(image_hdr_t);
    if (board_id_len > BOARD_ID_LEN)    /* Another sanity check */
        return -1;
    memset(board_id, 0, sizeof(board_id));
    memcpy(board_id, header+1, board_id_len);
    if (verify_board_id(board_id))
        return -1;

    /* Verify kernel and rootfs checksums */
    /* Note: rootfs is before kernel in the image */
#ifdef _DEBUG
    printf("kernel @ 0x%lX, len 0x%lx, chksum 0x%lx \n", header_len, kernel_len, kernel_chksum);
#endif
    if (kernel_len != 0 &&
        verify_checksum(&image_buf[header_len + rootfs_len], kernel_len, kernel_chksum))
        return -1;
        
#ifdef _DEBUG
    printf("rootfs @ 0x%lX, len 0x%lx, chksum 0x%lx \n", header_len + kernel_len, rootfs_len, rootfs_chksum);
#endif
    if (rootfs_len != 0 &&
        verify_checksum(&image_buf[header_len], rootfs_len, rootfs_chksum))
        return -1;
    
    return 0;
}

int write_image_to_flash(unsigned long image_addr)
{
    unsigned char *image_buf = (unsigned char *)image_addr;
    image_hdr_t *header      = (image_hdr_t *)image_buf;
    unsigned long header_len;
    unsigned long kernel_len, kernel_chksum;
    unsigned long rootfs_len, rootfs_chksum;
    char command[100];
    
    header_len    = swap32(header->header_len);

    kernel_len    = swap32(header->kernel_len);
    kernel_chksum = swap32(header->kernel_chksum);

    rootfs_len    = swap32(header->rootfs_len);
    rootfs_chksum = swap32(header->rootfs_chksum);    

    if (kernel_len)
    {
        unsigned long kernel_start = get_part_addr(PART_KERNEL);
        unsigned long kernel_end   = kernel_start + get_part_size(PART_KERNEL);

        /* Erase kernel partition */
        sprintf(command, "erase 0x%lX 0x%lX", kernel_start, kernel_end - 1);
#ifdef _DEBUG
        printf("command = %s \n", command);
#endif
        run_command(command, 0);
        
        /* Write kernel image */
        sprintf(command, "cp.b 0x%lX 0x%lX 0x%lX", image_addr + header_len + rootfs_len,
                                                   kernel_start,
                                                   kernel_len);
#ifdef _DEBUG
        printf("command = %s \n", command);
#endif
        run_command(command, 0);

        /* Write kernel length & checksum */
        sprintf(command, "cp.l 0x%lX 0x%lX 1", &kernel_len, kernel_end - 8);
#ifdef _DEBUG
        printf("command = %s \n", command);
#endif
        run_command(command, 0);

        sprintf(command, "cp.l 0x%lX 0x%lX 1", &kernel_chksum, kernel_end - 4);
#ifdef _DEBUG
        printf("command = %s \n", command);
#endif
        run_command(command, 0);
    }

    if (rootfs_len)
    {
        unsigned long rootfs_start = get_part_addr(PART_ROOTFS);
        unsigned long rootfs_end   = rootfs_start + get_part_size(PART_ROOTFS);

        /* Erase rootfs partition */
        sprintf(command, "erase 0x%lX 0x%lX", rootfs_start, rootfs_end - 1);
#ifdef _DEBUG
        printf("command = %s \n", command);
#endif
        run_command(command, 0);
        
        /* Write image to rootfs */
        sprintf(command, "cp.b 0x%lX 0x%lX 0x%lX", image_addr + header_len,
                                                   rootfs_start,
                                                   rootfs_len);
#ifdef _DEBUG
        printf("command = %s \n", command);
#endif
        run_command(command, 0);

        /* Write rootfs length & checksum */
        sprintf(command, "cp.l 0x%lX 0x%lX 1", &rootfs_len, rootfs_end - 8);
#ifdef _DEBUG
        printf("command = %s \n", command);
#endif
        run_command(command, 0);
        
        sprintf(command, "cp.l 0x%lX 0x%lX 1", &rootfs_chksum, rootfs_end - 4);
#ifdef _DEBUG
        printf("command = %s \n", command);
#endif
        run_command(command, 0);
    }

    return 0;
}

void mdelay(int msec)
{
    int i;
    for (i=0; i<msec; i++)
        udelay(1000);
}

static int gpio_control(int gpio, int value)
{
    if (value)
        mvGppValueSet(0, 1L << gpio, 1L << gpio);
    else
        mvGppValueSet(0, 1L << gpio, 0);

    return 0;
}

int power_led_control(int on_off)
{
    /* Turn on green means turn off amber */
    //if (on_off)
    //    gpio_control(GPIO_LED_AMBER, 0);

    gpio_control(GPIO_LED_AMBER, on_off);
    return 0;
}

int diag_led_control(int on_off)
{
    /* Turn on amber means turn off green */
    if (on_off)
        gpio_control(GPIO_LED_GREEN, 0);

    gpio_control(GPIO_LED_AMBER, on_off);
    return 0;
}

int wlan_led_control(int on_off)
{
    gpio_control(GPIO_LED_WLAN , on_off);
    return 0;
}

int wan_led_control(int on_off)
{
    gpio_control(GPIO_LED_WAN , on_off);
    return 0;
}

int check_reset(void)
{
    unsigned long nvram_start, nvram_end;
    char command[100];
    int  i;

    /* Reset pin is active low */
    if (!mvGppValueGet(0, 1L << GPIO_RESET))
    {
        /* Change power LED to Amber */
        diag_led_control(1);

        /* 
         * Wait 5 sec to see if we need to load default
         */
        for (i=0; i<500; i++)
        {
            mdelay(10);
            if (mvGppValueGet(0, 1L << GPIO_RESET))
                return 0;
        }

        printf("load default!\n");

        nvram_start = get_part_addr(PART_NVRAM);
        nvram_end = nvram_start + get_part_size(PART_NVRAM);
        sprintf(command, "erase 0x%lX 0x%lX", nvram_start, nvram_end - 1);
        run_command(command, 0);

        /* Toggle Amber LED while reset button is still pressed */
        while (1)
        {
            if (mvGppValueGet(0, 1L << GPIO_RESET))
                break;
                    
            diag_led_control(0);
            mdelay(750);
            diag_led_control(1);
            mdelay(250);

        }
        
        diag_led_control(0);
        mdelay(1000);
        printf("Reboot now!\n");
        diag_led_control(1);
        run_command("boot", 0);
    }

    return 0;
}
